@@ -1,6 +1,7 @@
 from __future__ import unicode_literals
 
 import gzip
+import logging
 import mimetypes
 import os
 import posixpath
@@ -14,6 +15,7 @@ from urllib.parse import urljoin, urlsplit, urlunsplit
 
 import boto3
 from botocore.client import Config
+from boto3.s3.transfer import TransferConfig
 from botocore.exceptions import ClientError
 from django.conf import settings
 from django.contrib.staticfiles.storage import ManifestFilesMixin
@@ -22,9 +24,11 @@ from django.core.files.base import File
 from django.core.files.storage import Storage
 from django.core.signals import setting_changed
 from django.utils.deconstruct import deconstructible
-from django.utils.encoding import filepath_to_uri, force_bytes, force_str, force_text
+from django.utils.encoding import filepath_to_uri, force_bytes, force_str
 from django.utils.timezone import make_naive, utc
 
+log = logging.getLogger(__name__)
+
 
 def _wrap_errors(func):
     @wraps(func)
@@ -36,7 +40,7 @@ def _wrap_errors(func):
             err_cls = OSError
             if code == "NoSuchKey":
                 err_cls = FileNotFoundError
-            raise err_cls("S3Storage error at {!r}: {}".format(name, force_text(ex)))
+            raise err_cls("S3Storage error at {!r}: {}".format(name, force_str(ex)))
     return _do_wrap_errors
 
 
@@ -44,10 +48,6 @@ def _callable_setting(value, name):
     return value(name) if callable(value) else value
 
 
-def _temporary_file():
-    return SpooledTemporaryFile(max_size=1024*1024*10)  # 10 MB.
-
-
 def _to_sys_path(name):
     return name.replace("/", os.sep)
 
@@ -98,7 +98,7 @@ class _Local(local):
     Thread-local connection manager.
 
     Boto3 objects are not thread-safe.
-    http://boto3.readthedocs.io/en/latest/guide/resources.html#multithreading-multiprocessing
+    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/resources.html#multithreading-and-multiprocessing
     """
 
     def __init__(self, storage):
@@ -150,7 +150,8 @@ class S3Storage(Storage):
         "AWS_S3_KMS_ENCRYPTION_KEY_ID": "",
         "AWS_S3_GZIP": True,
         "AWS_S3_SIGNATURE_VERSION": "s3v4",
-        "AWS_S3_FILE_OVERWRITE": False
+        "AWS_S3_FILE_OVERWRITE": False,
+        "AWS_S3_USE_THREADS": True,
     }
 
     s3_settings_suffix = ""
@@ -180,12 +181,17 @@ class S3Storage(Storage):
         if not self.settings.AWS_S3_BUCKET_NAME:
             raise ImproperlyConfigured(f"Setting AWS_S3_BUCKET_NAME{self.s3_settings_suffix} is required.")
         if self.settings.AWS_S3_PUBLIC_URL and self.settings.AWS_S3_BUCKET_AUTH:
-            raise ImproperlyConfigured(
-                f"Cannot use AWS_S3_BUCKET_AUTH{self.s3_settings_suffix} "
-                f"with AWS_S3_PUBLIC_URL{self.s3_settings_suffix}."
+            log.warning(
+                "Using AWS_S3_BUCKET_AUTH%s with AWS_S3_PUBLIC_URL%s. "
+                "Private files on S3 may be inaccessible via the public URL. "
+                "See https://github.com/etianen/django-s3-storage/issues/114 ",
+                self.s3_settings_suffix,
+                self.s3_settings_suffix,
             )
         # Create a thread-local connection manager.
         self._connections = _Local(self)
+        # Set transfer config for S3 operations
+        self._transfer_config = TransferConfig(use_threads=self.settings.AWS_S3_USE_THREADS)
 
     @property
     def s3_connection(self):
@@ -263,6 +269,10 @@ class S3Storage(Storage):
         # All done!
         return params
 
+    def new_temporary_file(self):
+        """Returns a new file to use when opening from or saving to S3"""
+        return SpooledTemporaryFile(max_size=1024*1024*10)  # 10 MB.
+
     @_wrap_errors
     def _open(self, name, mode="rb"):
         if mode != "rb":
@@ -270,7 +280,7 @@ class S3Storage(Storage):
         # Load the key into a temporary file. It would be nice to stream the
         # content, but S3 doesn't support seeking, which is sometimes needed.
         obj = self.s3_connection.get_object(**self._object_params(name))
-        content = _temporary_file()
+        content = self.new_temporary_file()
         shutil.copyfileobj(obj["Body"], content)
         content.seek(0)
         # Un-gzip if required.
@@ -288,7 +298,7 @@ class S3Storage(Storage):
         content.seek(0)
         # Convert content to bytes.
         if isinstance(content.file, TextIOBase):
-            temp_file = _temporary_file()
+            temp_file = self.new_temporary_file()
             temp_files.append(temp_file)
             for chunk in content.chunks():
                 temp_file.write(force_bytes(chunk))
@@ -305,7 +315,7 @@ class S3Storage(Storage):
             content_type_subtype = content_type_subtype.split("+")[-1]
             if content_type_family == "text" or content_type_subtype in ("xml", "json", "html", "javascript"):
                 # Compress the content.
-                temp_file = _temporary_file()
+                temp_file = self.new_temporary_file()
                 temp_files.append(temp_file)
                 with closing(gzip.GzipFile(name, "wb", 9, temp_file)) as gzip_file:
                     shutil.copyfileobj(content, gzip_file)
@@ -319,7 +329,8 @@ class S3Storage(Storage):
                 else:
                     content.seek(0)
         # Save the file.
-        self.s3_connection.put_object(Body=content.read(), **put_params)
+        self.s3_connection.upload_fileobj(content, put_params.pop('Bucket'), put_params.pop('Key'),
+                                          ExtraArgs=put_params, Config=self._transfer_config)
         # Close all temp files.
         for temp_file in temp_files:
             temp_file.close()
@@ -351,17 +362,32 @@ class S3Storage(Storage):
     def delete(self, name):
         self.s3_connection.delete_object(**self._object_params(name))
 
+    @_wrap_errors
+    def copy(self, src_name, dst_name):
+        self.s3_connection.copy_object(
+            CopySource=self._object_params(src_name),
+            **self._object_params(dst_name))
+
+    @_wrap_errors
+    def rename(self, src_name, dst_name):
+        self.copy(src_name, dst_name)
+        self.s3_connection.delete_object(**self._object_params(src_name))
+
     def exists(self, name):
         name = _to_posix_path(name)
         if name.endswith("/"):
             # This looks like a directory, but on S3 directories are virtual, so we need to see if the key starts
             # with this prefix.
-            results = self.s3_connection.list_objects_v2(
-                Bucket=self.settings.AWS_S3_BUCKET_NAME,
-                MaxKeys=1,
-                Prefix=self._get_key_name(name) + "/",  # Add the slash again, since _get_key_name removes it.
-            )
-            return "Contents" in results
+            try:
+                results = self.s3_connection.list_objects_v2(
+                    Bucket=self.settings.AWS_S3_BUCKET_NAME,
+                    MaxKeys=1,
+                    Prefix=self._get_key_name(name) + "/",  # Add the slash again, since _get_key_name removes it.
+                )
+            except ClientError:
+                return False
+            else:
+                return "Contents" in results
         # This may be a file or a directory. Check if getting the file metadata throws an error.
         try:
             self.meta(name)
@@ -498,6 +524,14 @@ class ManifestStaticS3Storage(ManifestFilesMixin, StaticS3Storage):
         "AWS_S3_MAX_AGE_SECONDS_CACHED": 60 * 60 * 24 * 365,  # 1 year.
     })
 
+    def _save(self, name, content):
+        # See: https://github.com/etianen/django-s3-storage/issues/141
+        # Fix adapted from: https://github.com/jschneier/django-storages/pull/968
+        content.seek(0)
+        with self.new_temporary_file() as tmp:
+            shutil.copyfileobj(content, tmp)
+            return super()._save(name, File(tmp))
+
     def post_process(self, *args, **kwargs):
         initial_aws_s3_max_age_seconds = self.settings.AWS_S3_MAX_AGE_SECONDS
         self.settings.AWS_S3_MAX_AGE_SECONDS = self.settings.AWS_S3_MAX_AGE_SECONDS_CACHED
